import json

from memory.error import ReservedAllocationWouldBeOver100PercentException
from memory.service import Service
from memory.memory import SystemMemory


class ConfigurationManager:

    def __init__(self, path, system_memory=SystemMemory()):
        self._path = path
        self._system_memory = system_memory

    def load(self):

        services = []

        with open(self._path) as config_file:

            config = json.load(config_file)

            if config is None:
                pass
            elif config['memory'] is None:
                pass
            elif config['memory']['allocation'] is None:
                pass
            elif config['memory']['allocation']['services'] is None:
                pass
            else:
                total_memory = self._system_memory.total

                for service in config['memory']['allocation']['services']:

                    name = service['name']
                    percent = float(service['percent'])

                    services.append(Service(name, percent, total_memory))

        allocation = AllocationConfiguration(services, self._system_memory)
        memory = MemoryConfiguration(allocation)
        config = Configuration(memory)

        return config

    def write(self, config, indent=4):

        with open(self._path, 'w') as config_file:
            config_output = config.to_json(indent)
            config_file.write(config_output)


class Configuration:

    def __init__(self, memory):
        self._memory = memory

    @property
    def memory(self):
        return self._memory

    def to_json(self, indent=4):
        return json.dumps(self.to_dict(), indent=indent)

    def to_dict(self):
        return {'memory': self._memory.to_dict()}


class MemoryConfiguration:
    
    def __init__(self, allocation):
        self._allocation = allocation

    @property
    def allocation(self):
        return self._allocation

    def to_dict(self):
        return {'allocation': self._allocation.to_dict()}


class AllocationConfiguration:
    
    def __init__(self, services=[], system_memory=SystemMemory()):
        self._services = services
        self._system_memory = system_memory

    @property
    def services(self):
        return self._services

    def add_service(self, service):
        self._services.append(service)

    def remove_service_by_name(self, name):
        self._services = [s for s in self.services if s.name != name]

    @property
    def total_allocation_percent(self):
        return sum([s.allocation_percent for s in self._services])

    def update_or_add_service(self, name, percent):

        existing_list = list(filter(lambda u:u.name == name, self._services))

        if len(existing_list) > 0:

            existing = existing_list[0]

            if self.total_allocation_percent - existing.allocation_percent + percent > 100.0:
                # Updated percent would push reserved percent over 100%
                raise ReservedAllocationWouldBeOver100PercentException()
            else:
                existing.allocation_percent = percent

        else:
            if self.total_allocation_percent + percent > 100.0:
                raise ReservedAllocationWouldBeOver100PercentException()
            else:
                self.add_service(Service(name, percent, self._system_memory.total))

    def to_dict(self):
        return {'services': [s.to_dict() for s in self._services]}