from client.client import Client
from references.error import VertexNotFoundException, EdgeNotDeletedException


class ReferenceManager:

    def __init__(self, session):
        self._session = session

    def find_all_vertices(self):

        vertices = []

        response = self._session.get('http://localhost:8082/api/endeavour/association/vertex').json()

        for json_vertex in response['vertices']:

            vertex = self._parse_vertex(json_vertex)

            vertices.append(vertex)

        return vertices

    def find_vertex(self, type, resource_id):

        url = 'http://localhost:8082/api/endeavour/association/vertex/{}/{}'.format(type, resource_id)

        response = self._session.get(url)

        return self._parse_vertex(response.json())

    def _parse_vertex(self, json_vertex):

        id = json_vertex['id']
        resource_id = json_vertex['resourceId']
        name = json_vertex['name']
        type = json_vertex['type']

        return Vertex(id, resource_id, name, type)

    def find_all_edges(self, type, resource_id):

        self._test_vertex_exists(type, resource_id)

        base_url = 'http://localhost:8082/api/endeavour/association/vertex/{}/{}'.format(type, resource_id)

        in_edges = self._find_edges('{}/edgein'.format(base_url))
        out_edges = self._find_edges('{}/edgeout'.format(base_url))

        return Edges(in_edges, out_edges)

    def _test_vertex_exists(self, type, resource_id):

        url = 'http://localhost:8082/api/endeavour/association/vertex/{}/{}'.format(type, resource_id)

        response = self._session.get(url)

        if not (response.status_code >= 200 and response.status_code < 300):
            raise VertexNotFoundException('Could not find vertex {}:{}'.format(type, resource_id))

    def _find_edges(self, url):

        edges = []

        edge_in_json = self._session.get(url).json()

        for json_edge in edge_in_json['edges']:

            edge = self._parse_edge(json_edge)

            edges.append(edge)

        return edges

    def _parse_edge(self, json_edge):

        label = json_edge['label']
        from_vertex = self._parse_vertex(json_edge['outVertex'])
        to_vertex = self._parse_vertex(json_edge['inVertex'])

        return Edge(label, from_vertex, to_vertex)

    def delete_vertex(self, type, resource_id):

        self._test_vertex_exists(type, resource_id)

        url = 'http://localhost:8082/api/endeavour/association/vertex/{}/{}'.format(type, resource_id)

        response = self._session.delete(url)

        if response.status_code == 404:
            raise VertexNotFoundException('Could not find vertex {}:{}'.format(type, resource_id))

        if response.status_code < 200 or response.status_code >= 300:
            raise VertexNotDeletedException("Failed to delete vertex. See SPP log for more information.")

    def delete_incoming_edges(self, type, resource_id, label):

        self._test_vertex_exists(type, resource_id)

        url = 'http://localhost:8082/api/endeavour/association/vertex/{}/{}/edgein/{}'.format(type, resource_id, label)

        response = self._session.delete(url)

        if response.status_code == 404:
            raise VertexNotFoundException('Could not find vertex {}:{}'.format(type, resource_id))

        if response.status_code < 200 or response.status_code >= 300:
            raise EdgeNotDeletedException("Failed to delete vertex. See SPP log for more information. Response code was {} and response body was: {}".format(response.status_code, response.json()))


class Vertex:

    def __init__(self, id, resource_id, name, type):
        self._id = id
        self._resource_id = resource_id
        self._name = name
        self._type = type

    @property
    def id(self):
        return self._id

    @property
    def resource_id(self):
        return self._resource_id

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type
        
  
class Edge:

    def __init__(self, label, from_vertex, to_vertex):
        self._label = label
        self._from_vertex = from_vertex
        self._to_vertex = to_vertex

    @property
    def label(self):
        return self._label

    @property
    def from_vertex(self):
        return self._from_vertex

    @property
    def to_vertex(self):
        return self._to_vertex
    
    
class Edges:

    def __init__(self, incoming, outgoing):
        self._incoming = incoming
        self._outgoing = outgoing

    @property
    def incoming(self):
        return self._incoming

    @property
    def outgoing(self):
        return self._outgoing
    
    