import click

from tabulate import tabulate
from getpass import getpass

from config.system import ConfigurationManager
from memory.memory import SystemMemory, ByteValue
from memory.service import ServiceFinder
from client.client import Client
from references.references import ReferenceManager


@click.group()
@click.pass_context
@click.option('--config', default='/opt/ECX/sys/config.json', help='The path to the configuration file')
def main(ctx, config):
    ctx.obj = {}
    ctx.obj['config_path'] = config


@main.group()
def memory():
    pass


@memory.group()
def system():
    pass


@memory.group()
def allocation():
    pass


@system.command('show')
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def show_system_memory(units, precision):

    sys_mem = SystemMemory()
    
    headers = ['Type', 'Value ({})'.format(units.upper()), '%']
    rows = [
        ('Total', display_units(sys_mem.total, units), 100),
        ('Available', display_units(sys_mem.available, units), sys_mem.available.percent),
        ('Used', display_units(sys_mem.used, units), sys_mem.used.percent),
        ('Free', display_units(sys_mem.available, units), sys_mem.free.percent),
        ('Active', display_units(sys_mem.active, units), sys_mem.active.percent),
        ('Inactive', display_units(sys_mem.inactive, units), sys_mem.inactive.percent),
        ('Buffers', display_units(sys_mem.buffers, units), sys_mem.buffers.percent),
        ('Cached', display_units(sys_mem.cached, units), sys_mem.cached.percent),     
        ('Shared', display_units(sys_mem.shared, units), sys_mem.shared.percent),
        ('Slab', display_units(sys_mem.slab, units), sys_mem.slab.percent),   
    ]

    click.echo('')
    title('System Memory')
    click.echo('')
    click.echo(tabulate(rows, headers=headers, floatfmt=".{}f".format(precision)))
    click.echo('')


def display_units(value, units):

    units_caseless = units.casefold()

    if units_caseless == 'kb'.casefold():
        return value.kilobytes
    elif units_caseless == 'mb'.casefold():
        return value.megabytes
    elif units_caseless == 'gb'.casefold():
        return value.gigabytes
    else:
        return value.bytes


def title(title):
    click.echo('=' * 80)
    click.echo(' {}'.format(title))
    click.echo('=' * 80)


def subtitle(title):
    click.echo('-' * 80)
    click.echo(' {}'.format(title))
    click.echo('-' * 80)


@allocation.command('show', help='Shows all memory allocation information')
@click.pass_context
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def show_memory_allocations(ctx, units, precision):
    
    sys_mem = SystemMemory()
    total_mem = sys_mem.total

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()

    total_allocation_bytes = sum([u.allocation.bytes for u in services])
    total_allocation = ByteValue(total_allocation_bytes, total_mem.bytes)
    unreserved = ByteValue(total_mem.bytes - total_allocation.bytes, total_mem.bytes)

    headers = ['Service', 'Memory ({})'.format(units.upper()), '%']
    rows = []

    reserved_headers = ['Category', 'Memory ({})'.format(units.upper()), '%']
    reserved_rows = [
        ('Allocated', display_units(total_allocation, units), total_allocation.percent),
        ('Unallocated', display_units(unreserved, units), unreserved.percent)
    ]

    for service in services:
        row = (service.name, display_units(service.allocation, units), service.allocation_percent)
        rows.append(row)
    
    click.echo('')
    title('Services')
    click.echo('')
    click.echo(tabulate(rows, headers=headers, floatfmt=".{}f".format(precision)))
    click.echo('')
    title('Allocations')
    click.echo('')
    click.echo(tabulate(reserved_rows, headers=reserved_headers, floatfmt=".{}f".format(precision)))
    click.echo('')
    

@allocation.command('get', help='Shows memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to obtain')
@click.option('--units', type=click.Choice(['b', 'kb', 'mb', 'gb'], case_sensitive=False), default='gb', help='Units to display')
@click.option('--precision', type=int, default=2, help='Decimal places to display')
def get_memory_allocation(ctx, service_name, units, precision):
    
    sys_mem = SystemMemory()

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    service_finder = ServiceFinder(config, sys_mem)

    service = service_finder.find_by_name(service_name)

    if service is not None:

        allocation = display_units(service.allocation, units)

        click.echo('{:.{prec}f}'.format(allocation, prec=precision))

    else:
        error('Cannot find service "{}"'.format(service_name))


def error(message):
    click.echo(message)
    exit(1)


@allocation.command('set', help='Configures memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to set')
@click.option('--percent', required=True, type=float, help='The percent of memory allocated to the service, as a decimal value')
def set_memory_allocation(ctx, service_name, percent):

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()
    
    sys_mem = SystemMemory()
    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()
    reserved_percent = sum([s.allocation_percent for s in services])

    try:
        config_manager = ConfigurationManager(ctx.obj['config_path'])
        config = config_manager.load()

        config.memory.allocation.update_or_add_service(service_name, percent)

        config_manager.write(config)

        click.echo('Set allocation for service "{}" to {}%'.format(service_name, percent))

    except NegativeAllocationPercentException as e:
        error('Cannot set allocation for service "{}" to less than 0%'.format(service_name))

    except ReservedAllocationWouldBeOver100PercentException as e:
        error('Cannot set allocation for service "{}" to {}%. Setting this value would set total allocations to over 100%'.format(service_name, percent))


@allocation.command('unset', help='Removes memory allocation information for a specific service')
@click.pass_context
@click.option('--service', 'service_name', required=True, help='The service whose memory allocation to unset')
def unset_memory_allocation(ctx, service_name):
    
    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    sys_mem = SystemMemory()
    service_finder = ServiceFinder(config, sys_mem)

    services = service_finder.find_all()
    reserved_percent = sum([s.allocation_percent for s in services])

    config_manager = ConfigurationManager(ctx.obj['config_path'])
    config = config_manager.load()

    config.memory.allocation.remove_service_by_name(service_name)

    config_manager.write(config)

    click.echo('Removed allocation for service "{}"'.format(service_name))


@main.group()
def references():
    pass


@references.group()
def vertex():
    pass


@vertex.command('show', help='Shows all reference vertices')
@click.pass_context
@click.option('--username', help='SPP username')
@click.option('--password', help='SPP password')
@click.option('--type', default=None, help='The type of the vertex to filter by')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def get_reference_vertex(ctx, username, password, type, name_length):

    try:
        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        vertices = ref_manager.find_all_vertices()

        headers = ['Name', 'ID', 'Type', 'Resource ID']
        rows = []

        if type is not None:
            vertices = filter(lambda v: v.type == type, vertices)

        for vertex in vertices:

            name = format_vertex_name(vertex.name, name_length)
            
            row = (name, vertex.id, vertex.type, vertex.resource_id)
            
            rows.append(row)
        
        click.echo('')
        title('Vertices')
        click.echo('')
        click.echo(tabulate(rows, headers=headers))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def obtain_credentials(cli_username, cli_password):

    username = cli_username
    password = cli_password

    if username is None or password is None:
        click.echo('')

    if username is None:
        username = input('  Enter SPP username: ')

    if password is None:
        password = getpass('  Enter SPP password: ')

    if username is None or password is None:
        click.echo('')

    return (username, password)


def format_vertex_name(name, name_length):

    if name is None:
        return '<unnamed>'

    elif len(name) > name_length:
        return name[:name_length] + '...'

    else:
        return name


@vertex.command('get')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def get__vertex(ctx, vertex, username, password, name_length):
    """
    Obtains the information for a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        (type, resource_id) = parse_vertex_name(vertex)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        vertex = ref_manager.find_vertex(type, resource_id)
        edges = ref_manager.find_all_edges(type, resource_id)

        incoming_edges_headers = [
            'Label\n\n', 
            'Outgoing Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        incoming_edges_rows = format_edge_rows(edges.incoming, name_length, lambda edge:edge.from_vertex)

        outgoing_edges_headers = [
            'Label\n\n', 
            'Incoming Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        outgoing_edges_rows = format_edge_rows(edges.outgoing, name_length, lambda edge:edge.to_vertex)

        click.echo('')
        title('Vertex Information')
        click.echo('')
        click.echo('  ID:          {}'.format(vertex.id))
        click.echo('  Name:        {}'.format(vertex.name))
        click.echo('  Type:        {}'.format(vertex.type))
        click.echo('  Resource ID: {}'.format(vertex.resource_id))
        click.echo('')
        title('Edges')
        click.echo('')
        subtitle('Incoming Edges')
        click.echo('')

        if len(incoming_edges_rows) > 0:
            click.echo(tabulate(incoming_edges_rows, headers=incoming_edges_headers))
        else:
            click.echo("None")
        
        click.echo('')
        click.echo('')
        subtitle('Outgoing Edges')
        click.echo('')

        if len(outgoing_edges_rows) > 0:
            click.echo(tabulate(outgoing_edges_rows, headers=outgoing_edges_headers))
        else:
            click.echo("None")
        
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def parse_vertex_name(name):
    
    split_name = name.split(':')
    
    type = split_name[0]
    resource_id = split_name[1]

    return (type, resource_id)


def format_edge_rows(edges, name_length, edge_direction_extractor):

    rows = []

    for edge in edges:

        vertex = edge_direction_extractor(edge)
        vertex_name = (vertex.name[:name_length] + '...') if len(vertex.name) > name_length else vertex.name
        
        row = (edge.label, vertex_name, vertex.id, vertex.type, vertex.resource_id)
        
        rows.append(row)

    return rows


@vertex.command('delete')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--force', '-f', is_flag=True, help="Forces the command the execute without confirmation")
def delete_reference_vertex(ctx, vertex, username, password, force):
    """
    Deletes a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        confirm(force, 'Vertex not deleted.')

        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        (type, resource_id) = parse_vertex_name(vertex)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.delete_vertex(type, resource_id)

        click.echo('')
        click.echo('  Successfully deleted vertex {}.'.format(vertex))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def confirm(force, message):

    if not force:

        click.echo('')
        confirmation = input('  Are you sure you want to perform this action? [Y/n] ')

        if not (confirmation == 'Y' or confirmation == 'y'):
            click.echo('')
            click.echo('  {}'.format(message))
            click.echo('')
            exit(0)


@references.group()
def edge():
    pass


@edge.group('in')
def edge_in():
    pass

@edge_in.command('delete')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--label', default="USES", help='The label of the edges to delete')
@click.option('--force', '-f', is_flag=True, help="Forces the command the execute without confirmation")
def delete_vertex_edges_in(ctx, vertex, username, password, label, force):
    """
    Deletes the incoming edges for a specific vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        confirm(force, 'Edges not deleted.')

        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        (type, resource_id) = parse_vertex_name(vertex)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.delete_incoming_edges(type, resource_id, label)

        click.echo('')
        click.echo('  Successfully deleted incoming edges with label {} for vertex {}.'.format(label, vertex))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


@vertex.command('show', help='Shows all reference vertices')
@click.pass_context
@click.option('--username', help='SPP username')
@click.option('--password', help='SPP password')
@click.option('--type', default=None, help='The type of the vertex to filter by')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def show_vertices(ctx, username, password, type, name_length):

    try:
        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        vertices = ref_manager.find_all_vertices()

        headers = ['Name', 'ID', 'Type', 'Resource ID']
        rows = []

        if type is not None:
            vertices = filter(lambda v: v.type == type, vertices)

        for vertex in vertices:

            name = format_vertex_name(vertex.name, name_length)
            
            row = (name, vertex.id, vertex.type, vertex.resource_id)
            
            rows.append(row)
        
        click.echo('')
        title('Vertices')
        click.echo('')
        click.echo(tabulate(rows, headers=headers))
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


def obtain_credentials(cli_username, cli_password):

    username = cli_username
    password = cli_password

    if username is None or password is None:
        click.echo('')

    if username is None:
        username = input('  Enter SPP username: ')

    if password is None:
        password = getpass('  Enter SPP password: ')

    if username is None or password is None:
        click.echo('')

    return (username, password)


@edge_in.command('get')
@click.pass_context
@click.argument('vertex')
@click.option('--username', default=None, help='SPP username')
@click.option('--password', default=None, help='SPP password')
@click.option('--name-length', default=80, help='The maximum character length of the vertex name')
def get_incoming_edges(ctx, vertex, username, password, name_length):
    """
    Obtains the information about incoming edges for a vertex.

    VERTEX is formatted as <type>:<resource-id>. For example: storageprofile:1000
    """

    try:
        (user_supplied_username, user_supplied_password) = obtain_credentials(username, password)
        (type, resource_id) = parse_vertex_name(vertex)
        
        client = Client()
        session = client.create_session(user_supplied_username, user_supplied_password)

        ref_manager = ReferenceManager(session)
        edges = ref_manager.find_all_edges(type, resource_id)

        incoming_edges_headers = [
            'Label\n\n', 
            'Outgoing Vertex\n\nName',
            '\n\nID',
            '\n\nType',
            '\n\nResource ID'
        ]
        incoming_edges_rows = format_edge_rows(edges.incoming, name_length)
        
        click.echo('')
        title('Incoming Edges')
        click.echo('')

        if len(incoming_edges_rows) > 0:
            click.echo(tabulate(incoming_edges_rows, headers=incoming_edges_headers))
        else:
            click.echo("None")
        
        click.echo('')

    except Exception as e:

        click.echo('')
        click.echo('  Error: {}'.format(str(e)))
        click.echo('')


if __name__ == '__main__':
    main()